// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build goexperiment.ms_tls_config_schannel

package tls

import (
	"internal/syscall/windows"
	"slices"
	"unsafe"
)

// init modifies the default cipher suites, the preference order, and the disabled cipher suites
// based on the cipher suites that Schannel supports.
//
// The user-visible behavior changes are:
// - The cipher suites used when [Config.CipherSuites] is nil, or when using TLS 1.3, only include those that Schannel supports.
// - The order in which cipher suites are tried, regardless of whether the user sets [Config.CipherSuites] or not, is the order that Schannel prefers.
func init() {
	cipherSuitesSchannel, err := cipherSuitesSchannel()
	if err != nil {
		panic(err) // This should never happen, as Schannel is always available on Windows.
	}

	// cipherSuitesSchannel contains the cipher suites that Schannel supports in its preference order.
	// We use this to filter the default cipher suites and to order the preference order.

	disableCipherSuites(CipherSuites(), cipherSuitesSchannel, disabledCipherSuites)
	disableCipherSuites(InsecureCipherSuites(), cipherSuitesSchannel, disabledCipherSuites)

	cipherSuitesPreferenceOrder = orderCipherSuites(cipherSuitesPreferenceOrder, cipherSuitesSchannel)
	cipherSuitesPreferenceOrderNoAES = orderCipherSuites(cipherSuitesPreferenceOrderNoAES, cipherSuitesSchannel)

	defaultCipherSuitesFIPS = filterCipherSuites(cipherSuitesSchannel, defaultCipherSuitesFIPS)
	defaultCipherSuitesTLS13 = filterCipherSuites(cipherSuitesSchannel, defaultCipherSuitesTLS13)
	defaultCipherSuitesTLS13NoAES = filterCipherSuites(cipherSuitesSchannel, defaultCipherSuitesTLS13NoAES)
	defaultCipherSuitesTLS13FIPS = filterCipherSuites(cipherSuitesSchannel, defaultCipherSuitesTLS13FIPS)
}

// cipherSuiteID returns the ID of the cipher suite with the given name.
func cipherSuiteID(name string) (uint16, bool) {
	for _, c := range CipherSuites() {
		if c.Name == name {
			return c.ID, true
		}
	}
	for _, c := range InsecureCipherSuites() {
		if c.Name == name {
			return c.ID, true
		}
	}
	return 0, false
}

// cipherSuitesSchannel returns all the cipher suites that Schannel supports in Schannel's preference order.
func cipherSuitesSchannel() ([]uint16, error) {
	// Get all the cipher suites that Schannel supports in preference order.
	var size uint32
	var funcs *windows.CRYPT_CONTEXT_FUNCTIONS
	err := windows.BCryptEnumContextFunctions(windows.CRYPT_LOCAL, unsafe.SliceData(windows.SSL_CONTEXT[:]), windows.NCRYPT_SCHANNEL_INTERFACE, &size, &funcs)
	if err != nil {
		return nil, err
	}
	defer windows.BCryptFreeBuffer(unsafe.Pointer(funcs))

	suites := make([]uint16, 0, funcs.Count) // order[i] will be the index of suites[i] in the Schannel list
	for i := range funcs.Count {
		name := windows.UTF16PtrToString(funcs.At(int(i)))
		id, ok := cipherSuiteID(name)
		if !ok {
			continue // cipher suite not found in the provided list
		}
		suites = append(suites, id)
	}
	return suites, nil
}

// filterCipherSuites filters the provided cipher suites, creating a new slice
// that only includes those that are in the allowed list.
func filterCipherSuites(suites, allowed []uint16) []uint16 {
	out := make([]uint16, 0, len(suites))
	for _, suite := range suites {
		if slices.Contains(allowed, suite) {
			out = append(out, suite)
		}
	}
	return out
}

// orderCipherSuites returns a new slice of cipher suites ordered according to the provided order.
// If suites contains a cipher suite that is not in the order, it will be placed
// at the end of the returned slice in the order they appear in suites.
func orderCipherSuites(suites []uint16, order []uint16) []uint16 {
	out := make([]uint16, 0, len(suites))
	for _, id := range order {
		if slices.Contains(suites, id) {
			out = append(out, id)
		}
	}
	for _, id := range suites {
		if !slices.Contains(out, id) {
			out = append(out, id)
		}
	}
	return out
}

// disableCipherSuites modifies the disabled map adding the cippher suites in supported
// that are not in allowed.
func disableCipherSuites(supported []*CipherSuite, allowed []uint16, disabled map[uint16]bool) {
	for _, suite := range supported {
		if !slices.Contains(allowed, suite.ID) {
			disabled[suite.ID] = true
		}
	}
}
